# %%
from pathlib import Path

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from iohub.ngff import open_ome_zarr
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import accuracy_score, classification_report, f1_score

from viscy.representation.embedding_writer import read_embedding_dataset

# %%
train_annotations = Path(
    "/hpc/projects/intracellular_dashboard/viral-sensor/2024_11_26_A549_ZIKA-sensor_ZIKV/3-phenotype/annotate-infection/combined_annotations.csv"
)
train_embeddings = Path(
    "/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/bootstrap-labels/generate-labels/sensor-2024-11-26.zarr"
)
val_annotations = Path(
    "/hpc/projects/intracellular_dashboard/viral-sensor/2024_08_14_ZIKV_pal17_48h/6-phenotype/combined_annotations.csv"
    # "/hpc/projects/intracellular_dashboard/viral-sensor/2024_11_05_A549_pAL10_24h/4-phenotype/annotate-infection/combined_annotations.csv"
)
val_embeddings = Path(
    "/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/bootstrap-labels/generate-labels/sensor-2024-08-14-annotation.zarr"
    # "/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/bootstrap-labels/generate-labels/sensor-2024-11-05.zarr"
)


# %%
def filter_train_fovs(fov_name: pd.Series) -> pd.Series:
    return fov_name.str[1:4].isin(["C/2", "B/3"])


def filter_val_fovs(fov_name: pd.Series) -> pd.Series:
    return fov_name.str[1:4].isin(["0/3"]) | (fov_name == "/0/4/000001")
    # return fov_name.isin(
    #     ["/0/15/000001", "/0/11/002000", "/0/11/002001", "/0/11/002002"]
    # )


def all_fovs(fov_name: pd.Series) -> pd.Series:
    return None


def load_features_and_annotations(embedding_path, annotation_path, filter_fn):
    dataset = read_embedding_dataset(embedding_path)
    features = dataset["features"][filter_fn(dataset["fov_name"])]
    annotation = pd.read_csv(annotation_path)
    annotation["fov_name"] = "/" + annotation["fov_name"]
    annotation = annotation.set_index(["fov_name", "id"])
    index = features["sample"].to_dataframe().reset_index(drop=True)[["fov_name", "id"]]
    selected = pd.merge(
        left=index, right=annotation, on=["fov_name", "id"], how="inner"
    )
    selected["infection_state"] = selected["infection_state"].astype("category")
    return features, selected["infection_state"], selected


# %%
train_features, train_annotation, train_selected = load_features_and_annotations(
    train_embeddings, train_annotations, filter_fn=filter_train_fovs
)
val_features, val_annotation, val_selected = load_features_and_annotations(
    val_embeddings, val_annotations, filter_fn=filter_val_fovs
)

model = LogisticRegression(class_weight="balanced", random_state=42, solver="liblinear")
model = model.fit(train_features, train_annotation)
train_prediction = model.predict(train_features)
val_prediction = model.predict(val_features)

print("Training\n", classification_report(train_annotation, train_prediction))
print("Validation\n", classification_report(val_annotation, val_prediction))

val_selected["label"] = val_selected["infection_state"].cat.codes
val_selected["prediction_binary"] = val_prediction

# %%
prediction = val_selected
prediction["HPI"] = prediction["t"] / 2 + 3
bins = [3, 6, 9, 12, 15, 18, 21, 24]
labels = [f"{start}-{end}" for start, end in zip(bins[:-1], bins[1:])]
prediction["stage"] = pd.cut(prediction["HPI"], bins=bins, labels=labels, right=True)
comparison = prediction.melt(
    id_vars=["fov_name", "id", "HPI"],
    value_vars=["label", "prediction_binary"],
    var_name="source",
    value_name="value",
)
with sns.axes_style("whitegrid"):
    ax = sns.lineplot(
        data=comparison,
        x="HPI",
        y="value",
        style="source",
        errorbar=None,
        color="gray",
    )
    ax.set_ylabel("Infection ratio")

# %%
accuracy_by_t = prediction.groupby(["stage"]).apply(
    lambda x: float(accuracy_score(x["label"], x["prediction_binary"]))
)
f1_by_t = prediction.groupby(["stage"]).apply(
    lambda x: float(f1_score(x["label"], x["prediction_binary"]))
)

metrics_df = pd.DataFrame(
    data={
        "accuracy": accuracy_by_t.values,
        "F1": f1_by_t.values,
    },
    index=f1_by_t.index,
).reset_index()

metrics_long = metrics_df.melt(
    id_vars=["stage"],
    value_vars=["accuracy"],
    var_name="metric",
    value_name="score",
)
with sns.axes_style("ticks"):
    plt.style.use("../figures/figure.mplstyle")
    g = sns.catplot(
        data=metrics_long,
        x="stage",
        y="score",
        kind="point",
        linewidth=1.5,
        linestyles="--",
        legend=False,
        color="gray",
    )
    g.set_axis_labels("HPI", "accuracy")
    g.figure.set_size_inches(3.5, 0.75)
    g.set(xlim=(-1, 7), ylim=(0.9, 1.0))
    plt.show()

# %%
g.savefig(
    Path.home()
    / "gdrive/publications/dynaCLR/2025_dynaCLR_paper/fig_manuscript_svg/figure_knowledge_distillation/figure_parts/teacher_accuracy.pdf",
    dpi=300,
    bbox_inches="tight",
)

# %%



from __future__ import annotations

import json
from pathlib import Path
from typing import TYPE_CHECKING

import torch
from monai.transforms import Compose, MapTransform, Transform
from pycocotools.coco import COCO
from tifffile import imread
from torch.utils.data import DataLoader, Dataset
from torchvision.ops import box_convert

from viscy.data.gpu_aug import GPUTransformDataModule
from viscy.data.typing import Sample

if TYPE_CHECKING:
    from multiprocessing.managers import DictProxy


class LiveCellDataset(Dataset):
    """
    LiveCell dataset.

    Parameters
    ----------
    images : list of Path
        List of paths to single-page, single-channel TIFF files.
    transform : Transform or Compose
        Transform to apply to the dataset.
    cache_map : DictProxy
        Shared dictionary for caching images.
    """

    def __init__(
        self,
        images: list[Path],
        transform: Transform | Compose,
        cache_map: DictProxy,
    ) -> None:
        self.images = images
        self.transform = transform
        self._cache_map = cache_map

    def __len__(self) -> int:
        return len(self.images)

    def __getitem__(self, idx: int) -> Sample:
        name = self.images[idx]
        if name not in self._cache_map:
            image = imread(name)[None, None]
            image = torch.from_numpy(image).to(torch.float32)
            self._cache_map[name] = image
        else:
            image = self._cache_map[name]
        sample = Sample(source=image)
        sample = self.transform(sample)
        if not isinstance(sample, list):
            sample = [sample]
        return sample


class LiveCellTestDataset(Dataset):
    """
    LiveCell dataset.

    Parameters
    ----------
    image_dir : Path
        Directory containing the images.
    transform : MapTransform | Compose
        Transform to apply to the dataset.
    annotations : Path
        Path to the COCO annotations file.
    load_target : bool, optional
        Whether to load the target images (default is False).
    load_labels : bool, optional
        Whether to load the labels (default is False).
    """

    def __init__(
        self,
        image_dir: Path,
        transform: MapTransform | Compose,
        annotations: Path,
        load_target: bool = False,
        load_labels: bool = False,
    ) -> None:
        self.image_dir = image_dir
        self.transform = transform
        self.coco = COCO(str(annotations))
        self.image_ids = list(self.coco.imgs.keys())
        self.load_target = load_target
        self.load_labels = load_labels

    def __len__(self) -> int:
        return len(self.image_ids)

    def __getitem__(self, idx: int) -> Sample:
        image_id = self.image_ids[idx]
        file_name = self.coco.imgs[image_id]["file_name"]
        image_path = self.image_dir / file_name
        image = imread(image_path)[None, None]
        image = torch.from_numpy(image).to(torch.float32)
        sample = Sample(source=image)
        if self.load_target:
            sample["target"] = image
        if self.load_labels:
            anns = self.coco.loadAnns(self.coco.getAnnIds(image_id)) or []
            boxes = [torch.tensor(ann["bbox"]).to(torch.float32) for ann in anns]
            masks = [
                torch.from_numpy(self.coco.annToMask(ann)).to(torch.bool)
                for ann in anns
            ]
            dets = {
                "boxes": box_convert(torch.stack(boxes), in_fmt="xywh", out_fmt="xyxy"),
                "labels": torch.zeros(len(anns)).to(torch.uint8),
                "masks": torch.stack(masks),
            }
            sample["detections"] = dets
            sample["file_name"] = file_name
        self.transform(sample)
        return sample


class LiveCellDataModule(GPUTransformDataModule):
    def __init__(
        self,
        train_val_images: Path | None = None,
        test_images: Path | None = None,
        train_annotations: Path | None = None,
        val_annotations: Path | None = None,
        test_annotations: Path | None = None,
        train_cpu_transforms: list[MapTransform] = [],
        val_cpu_transforms: list[MapTransform] = [],
        train_gpu_transforms: list[MapTransform] = [],
        val_gpu_transforms: list[MapTransform] = [],
        test_transforms: list[MapTransform] = [],
        batch_size: int = 16,
        num_workers: int = 8,
        pin_memory: bool = True,
    ) -> None:
        super().__init__()
        if train_val_images is not None:
            self.train_val_images = Path(train_val_images)
            if not self.train_val_images.is_dir():
                raise NotADirectoryError(str(train_val_images))
        if test_images is not None:
            self.test_images = Path(test_images)
            if not self.test_images.is_dir():
                raise NotADirectoryError(str(test_images))
        if train_annotations is not None:
            self.train_annotations = Path(train_annotations)
            if not self.train_annotations.is_file():
                raise FileNotFoundError(str(train_annotations))
        if val_annotations is not None:
            self.val_annotations = Path(val_annotations)
            if not self.val_annotations.is_file():
                raise FileNotFoundError(str(val_annotations))
        if test_annotations is not None:
            self.test_annotations = Path(test_annotations)
            if not self.test_annotations.is_file():
                raise FileNotFoundError(str(test_annotations))
        self._train_cpu_transforms = Compose(train_cpu_transforms)
        self._val_cpu_transforms = Compose(val_cpu_transforms)
        self._train_gpu_transforms = Compose(train_gpu_transforms)
        self._val_gpu_transforms = Compose(val_gpu_transforms)
        self.test_transforms = Compose(test_transforms)
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.pin_memory = pin_memory

    @property
    def train_cpu_transforms(self) -> Compose:
        return self._train_cpu_transforms

    @property
    def val_cpu_transforms(self) -> Compose:
        return self._val_cpu_transforms

    @property
    def train_gpu_transforms(self) -> Compose:
        return self._train_gpu_transforms

    @property
    def val_gpu_transforms(self) -> Compose:
        return self._val_gpu_transforms

    def setup(self, stage: str) -> None:
        if stage == "fit":
            self._setup_fit()
        elif stage == "test":
            self._setup_test()

    def _parse_image_names(self, annotations: Path) -> list[Path]:
        with open(annotations) as f:
            images = [f["file_name"] for f in json.load(f)["images"]]
        return sorted(images)

    def _setup_fit(self) -> None:
        cache_map = torch.multiprocessing.Manager().dict()
        train_images = self._parse_image_names(self.train_annotations)
        val_images = self._parse_image_names(self.val_annotations)
        self.train_dataset = LiveCellDataset(
            [self.train_val_images / f for f in train_images],
            transform=self.train_cpu_transforms,
            cache_map=cache_map,
        )
        self.val_dataset = LiveCellDataset(
            [self.train_val_images / f for f in val_images],
            transform=self.val_cpu_transforms,
            cache_map=cache_map,
        )

    def _setup_test(self) -> None:
        self.test_dataset = LiveCellTestDataset(
            self.test_images,
            transform=self.test_transforms,
            annotations=self.test_annotations,
            load_labels=True,
        )

    def test_dataloader(self) -> DataLoader:
        return DataLoader(
            self.test_dataset, batch_size=self.batch_size, num_workers=self.num_workers
        )



"""PCA and UMAP dimensionality reduction."""

import pandas as pd
import umap
from numpy.typing import NDArray
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
from xarray import Dataset


def compute_phate(
    embedding_dataset,
    n_components: int = 2,
    knn: int = 5,
    decay: int = 40,
    update_dataset: bool = False,
    **phate_kwargs,
) -> tuple[object, NDArray]:
    """
    Compute PHATE embeddings for features and optionally update dataset.

    Parameters
    ----------
    embedding_dataset : xarray.Dataset or NDArray
        The dataset containing embeddings, timepoints, fov_name, and track_id,
        or a numpy array of embeddings.
    n_components : int, optional
        Number of dimensions in the PHATE embedding, by default None
    knn : int, optional
        Number of nearest neighbors to use in the KNN graph, by default 5
    decay : int, optional
        Decay parameter for the Markov operator, by default 40
    update_dataset : bool, optional
        Whether to update the PHATE coordinates in the dataset, by default False
    phate_kwargs : dict, optional
        Additional keyword arguments for PHATE, by default None

    Returns
    -------
    tuple[object, NDArray]
        PHATE model and PHATE embeddings

    Raises
    ------
    ImportError
        If PHATE is not installed. Install with: pip install viscy[phate]
    """
    try:
        import phate
    except ImportError:
        raise ImportError(
            "PHATE is not available. Install with: pip install viscy[phate]"
        )

    # Get embeddings from dataset if needed
    embeddings = (
        embedding_dataset["features"].values
        if isinstance(embedding_dataset, Dataset)
        else embedding_dataset
    )

    # Compute PHATE embeddings
    phate_model = phate.PHATE(
        n_components=n_components, knn=knn, decay=decay, **phate_kwargs
    )
    phate_embedding = phate_model.fit_transform(embeddings)

    # Update dataset if requested
    if update_dataset and isinstance(embedding_dataset, Dataset):
        for i in range(
            min(2, phate_embedding.shape[1])
        ):  # Only update PHATE1 and PHATE2
            embedding_dataset[f"PHATE{i + 1}"].values = phate_embedding[:, i]

    return phate_model, phate_embedding


def compute_pca(embedding_dataset, n_components=None, normalize_features=True):
    """Compute PCA embeddings for features and optionally update dataset.

    Parameters
    ----------
    embedding_dataset : xarray.Dataset or NDArray
        The dataset containing embeddings, timepoints, fov_name, and track_id,
        or a numpy array of embeddings.
    n_components : int, optional
        Number of components to keep in the PCA, by default None
    normalize_features : bool, optional
        Whether to normalize the features, by default True

    Returns
    -------
    tuple[NDArray, pd.DataFrame]
        PCA embeddings and PCA DataFrame
    """

    embeddings = (
        embedding_dataset["features"].values
        if isinstance(embedding_dataset, Dataset)
        else embedding_dataset
    )

    if normalize_features:
        scaled_features = StandardScaler().fit_transform(embeddings)
    else:
        scaled_features = embeddings

    # Compute PCA with specified number of components
    PCA_features = PCA(n_components=n_components, random_state=42)
    pc_features = PCA_features.fit_transform(scaled_features)

    # Create base dictionary with id and fov_name
    if isinstance(embedding_dataset, Dataset):
        pca_dict = {
            "id": embedding_dataset["id"].values,
            "fov_name": embedding_dataset["fov_name"].values,
            "t": embedding_dataset["t"].values,
            "track_id": embedding_dataset["track_id"].values,
        }
    else:
        pca_dict = {}

    # Add PCA components for features
    for i in range(pc_features.shape[1]):
        pca_dict[f"PCA{i + 1}"] = pc_features[:, i]

    # Create DataFrame with all components
    pca_df = pd.DataFrame(pca_dict)

    return pc_features, pca_df


def _fit_transform_umap(
    embeddings: NDArray, n_components: int = 2, normalize: bool = True
) -> tuple[umap.UMAP, NDArray]:
    """Fit UMAP model and transform embeddings."""
    if normalize:
        embeddings = StandardScaler().fit_transform(embeddings)
    umap_model = umap.UMAP(n_components=n_components, random_state=42)
    umap_embedding = umap_model.fit_transform(embeddings)
    return umap_model, umap_embedding


def compute_umap(
    embedding_dataset: Dataset, normalize_features: bool = True
) -> tuple[umap.UMAP, umap.UMAP, pd.DataFrame]:
    """Compute UMAP embeddings for features and projections.

    Parameters
    ----------
    embedding_dataset : Dataset
        Xarray dataset with features and projections.
    normalize_features : bool, optional
        Scale the input to zero mean and unit variance before fitting UMAP,
        by default True

    Returns
    -------
    tuple[umap.UMAP, umap.UMAP, pd.DataFrame]
        UMAP models for features and projections,
        and DataFrame with UMAP embeddings
    """
    features = embedding_dataset["features"].values
    projections = embedding_dataset["projections"].values

    umap_features, umap_features_embedding = _fit_transform_umap(
        features, n_components=2, normalize=normalize_features
    )
    umap_projection, umap_projection_embedding = _fit_transform_umap(
        projections, n_components=2, normalize=normalize_features
    )

    # Prepare DataFrame with id and UMAP coordinates
    umap_df = pd.DataFrame(
        {
            "id": embedding_dataset["id"].values,
            "track_id": embedding_dataset["track_id"].values,
            "t": embedding_dataset["t"].values,
            "fov_name": embedding_dataset["fov_name"].values,
            "UMAP1": umap_features_embedding[:, 0],
            "UMAP2": umap_features_embedding[:, 1],
            "UMAP1_proj": umap_projection_embedding[:, 0],
            "UMAP2_proj": umap_projection_embedding[:, 1],
        }
    )

    return umap_features, umap_projection, umap_df



from concurrent.futures import ProcessPoolExecutor

import iohub.ngff as ngff
import numpy as np
import scipy.stats

import viscy.utils.image_utils as image_utils
import viscy.utils.masks as mask_utils


def mp_wrapper(fn, fn_args, workers):
    """Create and save masks with multiprocessing

    :param list of tuple fn_args: list with tuples of function arguments
    :param int workers: max number of workers
    :return: list of returned dicts from create_save_mask
    """
    with ProcessPoolExecutor(workers) as ex:
        # can't use map directly as it works only with single arg functions
        res = ex.map(fn, *zip(*fn_args))
    return list(res)


def mp_create_and_write_mask(fn_args, workers):
    """Create and save masks with multiprocessing. For argument parameters
    see mp_utils.create_and_write_mask.

    :param list of tuple fn_args: list with tuples of function arguments
    :param int workers: max number of workers
    :return: list of returned dicts from create_save_mask
    """
    with ProcessPoolExecutor(workers) as ex:
        # can't use map directly as it works only with single arg functions
        res = ex.map(create_and_write_mask, *zip(*fn_args))
    return list(res)


def add_channel(
    position: ngff.Position,
    new_channel_array,
    new_channel_name,
    overwrite_ok=False,
):
    """
    Adds a channels to the data array at position "position". Note that there is
    only one 'tracked' data array in current HCS spec at each position. Also
    updates the 'omero' channel-tracking metadata to track the new channel.

    The 'new_channel_array' must match the dimensions of the current array in
    all positions but the channel position (1) and have the same datatype

    Note: to maintain HCS compatibility of the zarr store, all positions (wells)
    must maintain arrays with congruent channels. That is, if you add a channel
    to one position of an HCS compatible zarr store, an additional channel must
    be added to every position in that store to maintain HCS compatibility.

    :param Position zarr_dir: NGFF position node object
    :param np.ndarray new_channel_array: array to add as new channel with matching
                            dimensions (except channel dim) and dtype
    :param str new_channel_name: name of new channel
    :param bool overwrite_ok: if true, if a channel with the same name as
                            'new_channel_name' is found, will overwrite
    """
    assert len(new_channel_array.shape) == len(position.data.shape) - 1, (
        "New channel array must match all dimensions of the position array, "
        "except in the inferred channel dimension: "
        f"array shape: {position.data.shape}, "
        "expected channel shape: "
        f"{(position.data.shape[0],) + position.data.shape[2:]}, "
        f"received channel shape: {new_channel_array.shape}"
    )
    # determine whether to overwrite or append
    if new_channel_name in position.channel_names and overwrite_ok:
        new_channel_index = list(position.channel_names).index(new_channel_name)
    else:
        new_channel_index = len(position.channel_names)
        position.append_channel(new_channel_name, resize_arrays=True)

    # replace or append channel
    position["0"][:, new_channel_index] = new_channel_array


def create_and_write_mask(
    position: ngff.Position,
    time_indices,
    channel_indices,
    structure_elem_radius,
    mask_type,
    mask_name,
    verbose=False,
):
    # TODO: rewrite docstring
    """
    Create mask *for all depth slices* at each time and channel index specified
    in this position, and save them both as an additional channel in the data array
    of the given zarr store and a separate 'untracked' array with specified name.
    If output_channel_index is specified as an existing channel index, will overwrite
    this channel instead.

    Saves custom metadata related to the mask creation in the well-level
    .zattrs in the 'mask' field.

    When >1 channel are used to generate the mask, mask of each channel is
    generated then added together. Foreground fraction is calculated on
    a timepoint-position basis. That is, it will be recorded as an average
    foreground fraction over all slices in any given timepoint.


    :param str zarr_dir: directory to HCS compatible zarr store for usage
    :param str position_path: path within store to position to generate masks for
    :param list time_indices: list of time indices for mask generation,
                            if an index is skipped over, will populate with
                            zeros
    :param list channel_indices: list of channel indices for mask generation,
                            if more than 1 channel specified, masks from all
                            channels are aggregated
    :param int structure_elem_radius: size of structuring element used for binary
                            opening. str_elem: disk or ball
    :param str mask_type: thresholding type used for masking or str to map to
                            masking function
    :param str mask_name: name under which to save untracked copy of mask in
                            position
    :param bool verbose: whether this process should send updates to stdout
    """

    shape = position.data.shape
    position_masks_shape = tuple([shape[0], len(channel_indices), *shape[2:]])

    # calculate masks over every time index and channel slice
    position_masks = np.zeros(position_masks_shape)
    position_foreground_fractions = {}

    for time_index in range(shape[0]):
        timepoint_foreground_fraction = {}

        for channel_index in channel_indices:
            channel_name = position.channel_names[channel_index]
            mask_array_chan_idx = channel_indices.index(channel_index)

            if "mask" in channel_name:
                print("\n")
                if mask_type in channel_name:
                    print(f"Found existing channel: '{channel_name}'.")
                    print("You are likely creating duplicates, which is bad practice.")
                print(f"Skipping mask channel '{channel_name}' for thresholding")
            else:
                # print progress update
                if verbose:
                    time_progress = f"time {time_index + 1}/{shape[0]}"
                    channel_progress = f"chan {channel_index}/{channel_indices}"
                    position_progress = f"pos {position.zgroup.name}"
                    p = (
                        f"Computing masks slice [{position_progress}, {time_progress},"
                        f" {channel_progress}]\n"
                    )
                    print(p)

                # get mask for image slice or populate with zeros
                if time_index in time_indices:
                    mask = get_mask_slice(
                        position_zarr=position.data,
                        time_index=time_index,
                        channel_index=channel_index,
                        mask_type=mask_type,
                        structure_elem_radius=structure_elem_radius,
                    )
                else:
                    mask = np.zeros(shape[-2:])

                position_masks[time_index, mask_array_chan_idx] = mask

                # compute & record channel-wise foreground fractions
                frame_foreground_fraction = float(
                    np.mean(position_masks[time_index, mask_array_chan_idx]).item()
                )
                timepoint_foreground_fraction[channel_name] = frame_foreground_fraction
        position_foreground_fractions[time_index] = timepoint_foreground_fraction

    # combine masks along channels and compute & record combined foreground fraction
    position_masks = np.sum(position_masks, axis=1)
    position_masks = np.where(position_masks > 0.5, 1, 0)
    for time_index in time_indices:
        frame_foreground_fraction = float(np.mean(position_masks[time_index]).item())
        timepoint_foreground_fraction["combined_fraction"] = frame_foreground_fraction

    # save masks as additional channel
    position_masks = position_masks.astype(position.data.dtype)
    new_channel_name = channel_name + "_mask"
    add_channel(
        position=position,
        new_channel_array=position_masks,
        new_channel_name=new_channel_name,
        overwrite_ok=True,
    )


def get_mask_slice(
    position_zarr,
    time_index,
    channel_index,
    mask_type,
    structure_elem_radius,
):
    """
    Given a set of indices, mask type, and structuring element,
    pulls an image slice from the given zarr array, computes the
    requested mask and returns.

    :param zarr.Array position_zarr: zarr array of the desired position
    :param time_index: see name
    :param channel_index: see name
    :param mask_type: see name,
                    options are {otsu, unimodal, mem_detection, borders_weight_loss_map}
    :param int structure_elem_radius: creation radius for the structuring
                    element
    :return np.ndarray mask: 2d mask for this slice
    """
    # read and correct/preprocess slice
    im = position_zarr[time_index, channel_index]
    im = image_utils.preprocess_image(im, hist_clip_limits=(1, 99))
    # generate mask for slice
    if mask_type == "otsu":
        mask = mask_utils.create_otsu_mask(im.astype("float32"))
    elif mask_type == "unimodal":
        mask = mask_utils.create_unimodal_mask(
            im.astype("float32"), structure_elem_radius
        )
    elif mask_type == "mem_detection":
        mask = mask_utils.create_membrane_mask(
            im.astype("float32"),
            structure_elem_radius,
        )
    elif mask_type == "borders_weight_loss_map":
        mask = mask_utils.get_unet_border_weight_map(im)
        mask = image_utils.im_adjust(mask).astype(position_zarr.dtype)

    return mask


def mp_get_val_stats(fn_args, workers):
    """
    Computes statistics of numpy arrays with multiprocessing

    :param list of tuple fn_args: list with tuples of function arguments
    :param int workers: max number of workers
    :return: list of returned df from get_im_stats
    """
    with ProcessPoolExecutor(workers) as ex:
        # can't use map directly as it works only with single arg functions
        res = ex.map(get_val_stats, fn_args)
    return list(res)


def get_val_stats(sample_values):
    """
    Computes the statistics of a numpy array and returns a dictionary
    of metadata corresponding to input sample values.

    :param list(float) sample_values: List of sample values at respective
                                        indices
    :return dict meta_row: Dict with intensity data for image
    """

    meta_row = {
        "mean": float(np.nanmean(sample_values)),
        "std": float(np.nanstd(sample_values)),
        "median": float(np.nanmedian(sample_values)),
        "iqr": float(scipy.stats.iqr(sample_values)),
    }
    return meta_row


def mp_sample_im_pixels(fn_args, workers):
    """Read and computes statistics of images with multiprocessing

    :param list of tuple fn_args: list with tuples of function arguments
    :param int workers: max number of workers
    :return: list of paths and corresponding returned df from get_im_stats
    """

    with ProcessPoolExecutor(workers) as ex:
        # can't use map directly as it works only with single arg functions
        res = ex.map(sample_im_pixels, *zip(*fn_args))
    return list(map(list, zip(*list(res))))


def sample_im_pixels(
    position: ngff.Position,
    grid_spacing,
    channel,
):
    # TODO move out of mp utils into normalization utils
    """
    Read and computes statistics of images for each point in a grid.
    Grid spacing determines distance in pixels between grid points
    for rows and cols.
    By default, samples from every time position and every z-depth, and
    assumes that the data in the zarr store is stored in [T,C,Z,Y,X] format,
    for time, channel, z, y, x.

    :param Position zarr_dir: NGFF position node object
    :param int grid_spacing: spacing of sampling grid in x and y
    :param int channel: channel to sample from

    :return list meta_rows: Dicts with intensity data for each grid point
    """
    image_zarr = position.data

    all_sample_values = []
    all_time_indices = list(range(image_zarr.shape[0]))
    all_z_indices = list(range(image_zarr.shape[2]))

    for time_index in all_time_indices:
        for z_index in all_z_indices:
            image_slice = image_zarr[time_index, channel, z_index, :, :]
            _, _, sample_values = image_utils.grid_sample_pixel_values(
                image_slice, grid_spacing
            )
            all_sample_values.append(sample_values)
    sample_values = np.stack(all_sample_values, 0).flatten()

    return position, sample_values



import logging
from typing import Literal, Sequence, TypedDict

import numpy as np
import torch
import torch.nn.functional as F
from lightning.pytorch import LightningModule
from pytorch_metric_learning.losses import NTXentLoss
from torch import Tensor, nn
from umap import UMAP

from viscy.data.typing import TrackingIndex, TripletSample
from viscy.representation.contrastive import ContrastiveEncoder
from viscy.utils.log_images import detach_sample, render_images

_logger = logging.getLogger("lightning.pytorch")


class ContrastivePrediction(TypedDict):
    features: Tensor
    projections: Tensor
    index: TrackingIndex


class ContrastiveModule(LightningModule):
    """Contrastive Learning Model for self-supervised learning."""

    def __init__(
        self,
        encoder: nn.Module | ContrastiveEncoder,
        loss_function: (
            nn.Module | nn.CosineEmbeddingLoss | nn.TripletMarginLoss | NTXentLoss
        ) = nn.TripletMarginLoss(margin=0.5),
        lr: float = 1e-3,
        schedule: Literal["WarmupCosine", "Constant"] = "Constant",
        log_batches_per_epoch: int = 8,
        log_samples_per_batch: int = 1,
        log_embeddings: bool = False,
        example_input_array_shape: Sequence[int] = (1, 2, 15, 256, 256),
    ) -> None:
        super().__init__()
        self.model = encoder
        self.loss_function = loss_function
        self.lr = lr
        self.schedule = schedule
        self.log_batches_per_epoch = log_batches_per_epoch
        self.log_samples_per_batch = log_samples_per_batch
        self.example_input_array = torch.rand(*example_input_array_shape)
        self.training_step_outputs = []
        self.validation_step_outputs = []
        self.log_embeddings = log_embeddings

    def forward(self, x: Tensor) -> tuple[Tensor, Tensor]:
        """Return both features and projections.

        Parameters
        ----------
        x : Tensor
            Input tensor

        Returns
        -------
        tuple[Tensor, Tensor]
            Tuple of (features, projections)
        """
        return self.model(x)

    def log_feature_statistics(self, embeddings: Tensor, prefix: str):
        mean = torch.mean(embeddings, dim=0).detach().cpu().numpy()
        std = torch.std(embeddings, dim=0).detach().cpu().numpy()
        _logger.debug(f"{prefix}_mean: {mean}")
        _logger.debug(f"{prefix}_std: {std}")

    def print_embedding_norms(self, anchor, positive, negative, phase):
        anchor_norm = torch.norm(anchor, dim=1).mean().item()
        positive_norm = torch.norm(positive, dim=1).mean().item()
        negative_norm = torch.norm(negative, dim=1).mean().item()
        _logger.debug(f"{phase}/anchor_norm: {anchor_norm}")
        _logger.debug(f"{phase}/positive_norm: {positive_norm}")
        _logger.debug(f"{phase}/negative_norm: {negative_norm}")

    def _log_metrics(
        self, loss, anchor, positive, stage: Literal["train", "val"], negative=None
    ):
        self.log(
            f"loss/{stage}",
            loss.to(self.device),
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
        )
        cosine_sim_pos = F.cosine_similarity(anchor, positive, dim=1).mean()
        euclidean_dist_pos = F.pairwise_distance(anchor, positive).mean()
        log_metric_dict = {
            f"metrics/cosine_similarity_positive/{stage}": cosine_sim_pos,
            f"metrics/euclidean_distance_positive/{stage}": euclidean_dist_pos,
        }

        if negative is not None:
            euclidean_dist_neg = F.pairwise_distance(anchor, negative).mean()
            cosine_sim_neg = F.cosine_similarity(anchor, negative, dim=1).mean()
            log_metric_dict[f"metrics/cosine_similarity_negative/{stage}"] = (
                cosine_sim_neg
            )
            log_metric_dict[f"metrics/euclidean_distance_negative/{stage}"] = (
                euclidean_dist_neg
            )

        self.log_dict(
            log_metric_dict,
            on_step=False,
            on_epoch=True,
            logger=True,
            sync_dist=True,
        )

    def _log_samples(self, key: str, imgs: Sequence[Sequence[np.ndarray]]):
        grid = render_images(imgs, cmaps=["gray"] * 3)
        self.logger.experiment.add_image(
            key, grid, self.current_epoch, dataformats="HWC"
        )

    def _log_step_samples(self, batch_idx, samples, stage: Literal["train", "val"]):
        """Common method for logging step samples"""
        if batch_idx < self.log_batches_per_epoch:
            output_list = (
                self.training_step_outputs
                if stage == "train"
                else self.validation_step_outputs
            )
            output_list.extend(detach_sample(samples, self.log_samples_per_batch))

    def log_embedding_umap(self, embeddings: Tensor, tag: str):
        _logger.debug(f"Computing UMAP for {tag} embeddings.")
        umap = UMAP(n_components=2)
        embeddings_np = embeddings.detach().cpu().numpy()
        umap_embeddings = umap.fit_transform(embeddings_np)

        # Log UMAP embeddings to TensorBoard
        self.logger.experiment.add_embedding(
            umap_embeddings,
            global_step=self.current_epoch,
            tag=f"{tag}_umap",
        )

    def training_step(self, batch: TripletSample, batch_idx: int) -> Tensor:
        anchor_img = batch["anchor"]
        pos_img = batch["positive"]
        _, anchor_projection = self(anchor_img)
        _, positive_projection = self(pos_img)
        negative_projection = None
        if isinstance(self.loss_function, NTXentLoss):
            indices = torch.arange(
                0, anchor_projection.size(0), device=anchor_projection.device
            )
            labels = torch.cat((indices, indices))
            # Note: we assume the two augmented views are the anchor and positive samples
            embeddings = torch.cat((anchor_projection, positive_projection))
            loss = self.loss_function(embeddings, labels)
            self._log_step_samples(batch_idx, (anchor_img, pos_img), "train")
        else:
            neg_img = batch["negative"]
            _, negative_projection = self(neg_img)
            loss = self.loss_function(
                anchor_projection, positive_projection, negative_projection
            )
            self._log_step_samples(batch_idx, (anchor_img, pos_img, neg_img), "train")
        self._log_metrics(
            loss=loss,
            anchor=anchor_projection,
            positive=positive_projection,
            negative=negative_projection,
            stage="train",
        )
        return loss

    def on_train_epoch_end(self) -> None:
        super().on_train_epoch_end()
        self._log_samples("train_samples", self.training_step_outputs)
        # Log UMAP embeddings for validation
        if self.log_embeddings:
            embeddings = torch.cat(
                [output["embeddings"] for output in self.validation_step_outputs]
            )
            self.log_embedding_umap(embeddings, tag="train")
        self.training_step_outputs = []

    def validation_step(self, batch: TripletSample, batch_idx: int) -> Tensor:
        """Validation step of the model."""
        anchor = batch["anchor"]
        pos_img = batch["positive"]
        _, anchor_projection = self(anchor)
        _, positive_projection = self(pos_img)
        negative_projection = None
        if isinstance(self.loss_function, NTXentLoss):
            indices = torch.arange(
                0, anchor_projection.size(0), device=anchor_projection.device
            )
            labels = torch.cat((indices, indices))
            # Note: we assume the two augmented views are the anchor and positive samples
            embeddings = torch.cat((anchor_projection, positive_projection))
            loss = self.loss_function(embeddings, labels)
            self._log_step_samples(batch_idx, (anchor, pos_img), "val")
        else:
            neg_img = batch["negative"]
            _, negative_projection = self(neg_img)
            loss = self.loss_function(
                anchor_projection, positive_projection, negative_projection
            )
            self._log_step_samples(batch_idx, (anchor, pos_img, neg_img), "val")
        self._log_metrics(
            loss=loss,
            anchor=anchor_projection,
            positive=positive_projection,
            negative=negative_projection,
            stage="val",
        )
        return loss

    def on_validation_epoch_end(self) -> None:
        super().on_validation_epoch_end()
        self._log_samples("val_samples", self.validation_step_outputs)
        # Log UMAP embeddings for training
        if self.log_embeddings:
            embeddings = torch.cat(
                [output["embeddings"] for output in self.training_step_outputs]
            )
            self.log_embedding_umap(embeddings, tag="val")

        self.validation_step_outputs = []

    def configure_optimizers(self):
        optimizer = torch.optim.AdamW(self.parameters(), lr=self.lr)
        return optimizer

    def predict_step(
        self, batch: TripletSample, batch_idx, dataloader_idx=0
    ) -> ContrastivePrediction:
        """Prediction step for extracting embeddings."""
        features, projections = self.model(batch["anchor"])
        return {
            "features": features,
            "projections": projections,
            "index": batch["index"],
        }
